import logging
import os
import typing

from remotemanager.serialisation.sendablemixin import SendableMixin
from remotemanager.logging.utils import format_iterable
from remotemanager.utils.uuid import generate_uuid

from datetime import datetime


class Runner(SendableMixin):
    """
    The Runner class stores any info pertaining to this specific run. E.g.
    Arguments, result, run status, files, etc.
    """

    _defaults = {'skip': False}

    _state_direct_remote = 'disconnected - remote direct run'
    _state_script_remote = 'script submitted remotely'
    _state_script_local = 'script submitted locally'

    def __init__(self,
                 arguments: dict,
                 **kwargs):

        self._logger = logging.getLogger(__name__ + '.Runner')

        self._run_options = self._set_defaults(kwargs)

        self._args = arguments
        self._uuid = generate_uuid(format_iterable(arguments))
        self._time_format = '%Y-%m-%d %H:%M:%S'
        self._state_time = None
        self._state = None
        self._result = None

        self._logger.info(f'new runner (id {self.uuid}) created')

        self._history = {}
        self.state = 'created'

    @staticmethod
    def _set_defaults(kwargs: dict = None) -> dict:
        """
        Sets default arguments as expected. If used as a staticmethod, returns
        the defaults
        """

        if kwargs is None:
            kwargs = {}

        for k, v in Runner._defaults.items():
            if k not in kwargs:
                kwargs[k] = v

        return kwargs

    def _parse_dir(self,
                   dir: str) -> str:
        """
        ensure a properly formatted local path for this run

        Args:
            dir (str):
                path

        Returns:
            formatted abspath
        """
        if dir is None:
            return os.getcwd()
        elif dir == '':
            return os.getcwd()
        elif os.path.isabs(dir):
            return dir
        return os.path.abspath(dir)

    @property
    def uuid(self):
        """
        The uuid of this runner
        """
        return self._uuid

    @property
    def short_uuid(self):
        """
        A short uuid for filenames
        """
        return self.uuid[:8]

    @property
    def runfile(self):
        """
        Runfile name
        """
        return f'{self.short_uuid}-run.py'

    @property
    def runpath(self):
        """
        Runfile name
        """
        return os.path.join(self.local_dir, f'{self.short_uuid}-run.py')

    @property
    def local_dir(self):
        """
        Local staging directory
        """
        return self._run_options.get('local_dir', 'staging')

    @local_dir.setter
    def local_dir(self,
                  path: str) -> None:
        self._run_options['local_dir'] = path

    @property
    def run_dir(self):
        """
        Intended running directory. If not set, attempts to use remote_dir and
        then falls back on local_dir
        """
        if 'run_dir' in self._run_options:
            return self._run_options['run_dir']
        return self._run_options.get('remote_dir', self.local_dir)

    @run_dir.setter
    def run_dir(self,
                dir: str) -> None:
        self._run_options['run_dir'] = dir

    @property
    def resultfile(self):
        """
        Result file name
        """
        return f'{self.short_uuid}-result.yaml'

    @property
    def resultpath(self):
        """
        Path to result file
        Currently a concatenation of the local dir and file name
        """
        # TODO change this to an adaptive path
        return os.path.join(self.run_dir, self.resultfile)

    @property
    def local_resultpath(self):
        """
        Path to local version of result file
        """
        # TODO change this to an adaptive path
        return os.path.join(self.local_dir, self.resultfile)

    @property
    def args(self):
        """
        Arguments for the function
        """
        return self._args

    @property
    def result(self):
        """
        Result (If available)
        """
        return self._result

    @result.setter
    def result(self, result) -> None:
        self._result = result
        self.state = 'completed'

    @property
    def state(self):
        return self._state

    @state.setter
    def state(self,
              newstate: str) -> None:
        """
        Update the state and store within the runner history
        """

        state_time = datetime.now()

        self.insert_history(state_time, newstate)

        self._state_time = int(state_time.strftime('%s'))
        self._state = newstate

    def format_time(self, time):
        return time.strftime(self._time_format)

    @property
    def history(self):
        """
        Returns (dict):
            State history of this runner
        """
        return self._history

    @property
    def status_list(self):
        return list(self._history.values())

    def insert_history(self,
                       time: datetime,
                       newstate: str) -> None:
        """
        Insert a state into this runner's history
        Args:
            time (datetime.time):
                time this state change occurred
            newstate (str):
                status to update

        Returns:
            None
        """
        if not isinstance(time, datetime):
            raise ValueError(f'time of type {type(time)} should be a datetime '
                             f'instance')

        timekey = self.format_time(time)
        if timekey not in self._history:
            self._logger.info(f'updating runner {self.short_uuid} state -> '
                              f'{newstate}')
            self._history[timekey] = newstate
        else:
            self._logger.info(f'appending runner {self.short_uuid} state -> '
                              f'{newstate}')
            self._history[timekey] += f' + {newstate}'

    @property
    def last_updated(self):
        """
        Time that this runner state last changed
        """
        return self._state_time

    def run(self,
            function: typing.Callable,
            **kwargs) -> None:
        """
        Substitutes runner args into function and runs, storing the result
        Args:
            function:
                typing.Callable:
                    function to run
        Returns:
            None
        """
        self._logger.info(f'running function {self.uuid}')
        self.state = 'submitted'

        self.update_run_options(kwargs)

        result = function(**self.args)

        self.result = result
        self._logger.info(f'run complete for {self.uuid}')

    @property
    def is_finished(self):
        """
        Attempts to determine if this runner has completed its run

        Returns (bool):
            completion status
        """
        self._logger.info(f'checking finished state of runner '
                          f'{self.short_uuid}')

        if self.state == 'completed':
            self._logger.info('state is completed: True')
            return True

        elif self.state == Runner._state_script_local:
            self._logger.info(f'script based submission, checking for '
                              f'file {self.resultpath}')
            # last status update was a script based submission
            # check for results file
            return os.path.isfile(self.resultpath)

        elif self.state == Runner._state_script_remote:
            self._logger.info('remote scripted run, dataset needs '
                              'to check this')
            # need to search for the results file
            return False

        elif self.state == Runner._state_direct_remote:
            self._logger.info('remote direct run, dataset needs '
                              'to check this')
            # this runner is "orphaned", and the result needs to be grabbed
            # remotely
            return False

        return False

    def update_run_options(self,
                           run_args: dict) -> None:
        """
        Update run args with dict `run_args`
        Args:
            run_args (dict):
                new run arguments

        Returns:
            None
        """
        self._logger.info('updating run options with new run args:')
        self._logger.info(format_iterable(run_args))

        self._run_options.update(run_args)

    def run_option(self,
                   option: str,
                   default=None):
        """
        Return a run option

        Args:
            option (str):
                key to search for
            default:
                default argument to provide to get

        Returns:
            option if available, else None
        """
        ret = self._run_options.get(option, default)
        self._logger.debug(f'getting run option {option}: {ret}')
        return ret

    @property
    def run_args(self):
        """
        Display the run arguments

        Returns:
            (dict) run_args
        """
        return self._run_options



import yaml
import logging
import os
import typing
from datetime import datetime

from remotemanager.connection.url import URL
from remotemanager.storage.database import Database
from remotemanager.serialisation.function import Function
from remotemanager.dataset.runner import Runner
import remotemanager.transport as tp
from remotemanager.transport.rsync import rsync
from remotemanager.transport.cp import cp
from remotemanager.serialisation.sendablemixin import SendableMixin
from remotemanager.utils.uuid import generate_uuid
from remotemanager.logging.utils import format_iterable

_logger = logging.getLogger(__name__)


class Dataset(SendableMixin):
    """
    Bulk holder for remote runs. The Dataset class handles anything regarding
    the runs as a group. Running, retrieving results, sending to remote, etc.

    Args:
        function (Callable):
            function to run
        url (URL):
            connection to remote (optional)
        transport (tp.transport.Transport):
            transport system to use, if a specific is required. Defaults to
            transport.rsync
        dbfile (str):
            filename for the database file to be associated with this dataset
        script (str):
            callscript required to run the jobs in this dataset
        submitter (str):
            command to exec any scripts with. Defaults to "bash"
        name (str):
            optional name for this dataset. Will be used for runscripts
        global_run_args:
            any further (unchanging) arguments to be passed to the runner(s)
    """

    # DEV NOTE: arguments must be None for computer-url override to function
    def __init__(self,
                 function: typing.Callable,
                 url: URL = None,
                 transport: tp.transport.Transport = None,
                 dbfile: str = None,
                 script: str = None,
                 submitter: str = None,
                 name: str = None,
                 skip: bool = None,
                 **global_run_args):

        self._logger = logging.getLogger(__name__ + '.Dataset')
        self._logger.info('initialising Database')
        try:
            self._function = Function(function)
        except TypeError:
            _logger.info('dummy call detected, exiting early')
            return

        self._global_run_args = global_run_args

        # dataset uuid is equal to Function uuid for now
        self._name = name if name is not None else 'dataset'
        self._uuid = generate_uuid(self._function.uuid + 'Dataset' + self.name)

        self._defaultfile = f'dataset-file-{self.short_uuid}.yaml'
        if dbfile is None:
            dbfile = os.path.join(os.getcwd(), self._defaultfile)

        self._dbfile = dbfile
        self._script = script if script is not None else '#!/bin/bash'
        self._submitter = submitter if submitter is not None else 'bash'
        self._scriptfile = f'run-{self.name}.sh'
        self.skip = skip if skip is not None else False

        self._url = None
        self._transport = None
        self.url = url
        self.transport = transport

        if os.path.isfile(self._dbfile):
            self._logger.info(f'unpacking database from {self._dbfile}')

            # create a "temporary" database from the found file
            self._database = Database(self._dbfile)
            # update it with any new values
            self.database.update(self.pack())
            # unpack from here to retrieve
            payload = self.database._storage[self.uuid]
            self.inject_payload(payload)

        else:
            self._runs = {}
            self._results = []

        # database property creates the database if it does not exist
        self.database.update(self.pack())

    @classmethod
    def recreate(cls,
                 dbfile: str,
                 function = None,
                 uuid: str = None):
        """
        Re-create the dataset from a function or uuid

        Args:
            function (Callable):
                If a dataset exists for this function, re-create it
            uuid (str):
                search the database for a uuid, and re-create the object

        Returns:
            Dataset
        """
        if function is None and uuid is None:
            raise ValueError('please provide a function or uuid to search for')

        if not os.path.isfile(dbfile):
            raise FileNotFoundError(f'could not find file {dbfile}')

        database = Database(file=dbfile)

        if uuid is None:
            # create a uuid if not otherwise found
            uuid = generate_uuid(Function(function).uuid)

        reinit = database.find(uuid)

        return cls.unpack(reinit)

    def __getattribute__(self, item):
        """
        Redirect Dataset.attribute calls to _global_run_args if possible.
        Allows for run global_run_args to be kept seperate
        Args:
            item:
                attribute to fetch
        """
        # TODO: keep an eye on this, it's hacky and liable to break
        if item != '_global_run_args' \
                and hasattr(self, '_global_run_args') \
                and item in self._global_run_args:
            return self._global_run_args.get(item)
        return object.__getattribute__(self, item)

    def __eq__(self, other):
        if not isinstance(other, type(self)):
            raise ValueError(f'Cannot compare Dataset against type '
                             f'{type(other)}')
        return self.uuid == other.uuid

    @property
    def database(self):
        """
        Access to the stored database object.
        Creates a connection if none exist.

        Returns:
            Database
        """
        if not hasattr(self, '_database'):
            self._database = Database(file=self._dbfile)
        return self._database
    
    def pack(self, **kwargs):
        if len(kwargs) == 0:
            self._logger.info('Dataset override pack called')
        else:
            self._logger.info('Data override pack called with kwargs')
            self._logger.info(f'{format_iterable(kwargs)}')
        return super().pack(uuid=self._uuid, **kwargs)

    def set_run_option(self, key, val):
        """
        Update a glopal run option `key` with value `val`
        Args:
            key (str):
                option to be updated
            val:
                value to set

        """
        self._global_run_args[key] = val

    def append_run(self,
                   args = None,
                   arguments = None,
                   **run_args):
        """
        Serialise arguments for later runner construction

        Args:
            args (dict):
                dictionary of arguments to be unpacked
            arguments(dict):
                alias for args
        """
        if args is None and arguments is not None:
            args = arguments

        # first grab global arguments and update them with explicit args
        run_arguments = {k: v for k, v in self._global_run_args.items()}
        run_arguments.update(run_args)

        tmp = Runner(args,
                     **run_arguments)

        if tmp.uuid not in self._runs:
            self._logger.debug(f'appending new run with uuid {tmp.uuid}')
            self._runs[tmp.uuid] = tmp
        else:
            self._logger.debug(f'runner with uuid {tmp.uuid} already exists')

        self.database.update(self.pack())

    @property
    def runners(self):
        return self._runs

    @property
    def runner_list(self):
        return [run for run in self._runs.values()]

    @property
    def function(self):
        return self._function

    @property
    def global_run_args(self):
        return self._global_run_args

    @property
    def script(self):
        return self._script

    @script.setter
    def script(self, script):
        self._script = script

    @property
    def submitter(self):
        return self._submitter

    @submitter.setter
    def submitter(self, submitter):
        self._submitter = submitter

    @property
    def url(self):
        if not hasattr(self, '_url'):
            self.url = None
        return self._url

    @url.setter
    def url(self, url=None):
        """
        Returns an empty (local) url connection if url is None

        Args:
            url (URL):
                url to be verified
        """
        self._logger.info(f'new url is being set to {url}')
        if url is None:
            self._logger.warning('no URL specified for this dataset, creating '
                                 'localhost')
            self._url = URL()
        else:
            self._url = url

        if not type(url) == URL and issubclass(type(url), URL):

            init_attrs = locals()

            self._logger.info('url is a computer, collecting attributes')

            self._script = url.script

            attrs = ['name', 'script', 'submitter']

            for attr in attrs:

                init_set = init_attrs.get(attr, None)

                self._logger.debug(f'attr {attr} is set to {init_set}')
                if init_set is None:
                    preset = getattr(url, attr, None)
                    setattr(self, attr, preset)

                    self._logger.info(f'set attribute {attr} to {preset}')

        timeout = self._global_run_args.get('timeout', None)
        max_timeouts = self._global_run_args.get('max_timeouts', None)

        self._url.timeout = timeout
        self._url.max_timeouts = max_timeouts

    @property
    def transport(self):
        if not hasattr(self, '_transport'):
            self.transport = None
        return self._transport

    @transport.setter
    def transport(self, transport=None):
        if transport is None:
            self._logger.warning('no transport specified for this dataset, '
                                 'creating basic rsync')
            self._transport = tp.rsync(self.url)
        else:
            self._transport = transport

    @property
    def dbfile(self):
        return self.database.path

    def remove_database(self):
        os.remove(self.dbfile)

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, name):
        if not isinstance(name, str):
            try:
                name = str(name)
            except TypeError:
                raise ValueError('name can only be str type')

        self._name = name

    @property
    def uuid(self):
        return self._uuid

    @property
    def short_uuid(self):
        return self._uuid[:8]

    def set_all_runner_states(self, state):
        for runner in self.runner_list:
            runner.state = state

    def check_all_runner_states(self, state):
        return all([r.state == state for r in self.runner_list])

    def run(self,
            **run_args):
        """
        Run the functions

        Args:
            run_args:
                any arguments to pass to the runners during this run.
                will override any "global" arguments set at Dataset init
        """

        if self.skip:
            self._logger.warning('running with skip = True, '
                                 'checking runner archive')

            for runner in self.runner_list:
                uuid = runner.uuid

                payload = self.database.find(uuid)
                archived = Runner.unpack(payload)
                self._logger.info(f'updating runner {runner.short_uuid} with '
                                  f'archived result')
                if 'completed' in archived.status_list:
                    runner = archived
                    runner.status = 'collected archived completion'

        if hasattr(self.url, 'script'):
            self.script = self.url.script

        for arg, val in run_args.items():
            if arg not in self._global_run_args:
                self._logger.debug(f'updating global arg {arg} -> {val}')
                self._global_run_args[arg] = val

        if self.script:
            self._run_scripted(run_args)
        elif not self.url.is_local:
            self._run_direct_remote(run_args)
        else:
            self._run_direct_local(run_args)

        self.database.update(self.pack())

    def _run_scripted(self, run_args):

        self._logger.info('preparing for scripted run')

        script = [self.script]

        asynchronous = getattr(self, 'asynchronous', False)

        for runner in self.runner_list:
            if 'completed' in runner.status_list and self.skip:
                msg = f'skipping run for runner ' \
                      f'{runner.short_uuid}'
                self._logger.info(msg)
                print(msg)
                continue

            runner.update_run_options(run_args)
            runner.state = runner._state_script_remote
            staged = os.path.join(runner.local_dir, runner.runfile)

            if not os.path.isdir(runner.local_dir):
                self._logger.info('creating local dir for the runner')
                os.makedirs(runner.local_dir)

            # TODO convert to full serialisation
            runfile = ['import yaml',
                       'import os',
                       self.function.dump_to_string(runner.args),
                       f'\twith open("{runner.resultfile}", "w+") as o:',
                       '\t\tyaml.dump(result, o)']

            with open(staged, 'w+') as o:
                o.write('\n'.join(runfile))

            rundir = runner.run_dir
            if self.url.is_local:
                rundir = os.path.join(os.getcwd(), rundir)
            else:
                rundir = f'$HOME/{rundir}'
            run_line = f'cd "{rundir}" && '\
                       f'touch "{runner.runfile}" && '\
                       f'{self.url.python} "{runner.runfile}"'

            asynchronous = runner.run_option('asynchronous', True)
            if asynchronous and self.submitter == 'bash':
                self._logger.debug('appending "&" for async run')
                run_line += ' &'

            script.append(run_line)

            self.transport.queue_for_push(runner.runfile,
                                          runner.local_dir,
                                          runner.run_dir)

        _scriptfile = os.path.join(runner.local_dir, self._scriptfile)
        self._logger.info(f'writing scriptfile into {_scriptfile}')
        with open(_scriptfile, 'w+') as o:
            o.write('\n'.join(script))

        self._logger.info(f'queuing run script into folder {runner.run_dir}')
        self.transport.queue_for_push(os.path.split(_scriptfile)[1],
                                      os.path.split(_scriptfile)[0],
                                      runner.run_dir)

        self.transport.transfer()

        self._logger.debug('running the script')

        remote_script_file = os.path.join(runner.run_dir, self._scriptfile)
        self.url.cmd(f'{self.submitter} {remote_script_file}',
                     asynchronous=asynchronous)

    def _run_direct_local(self, run_args):

        self._logger.info('running local direct')
        for runner in self.runner_list:
            if 'completed' in runner.status_list and self.skip:
                msg = f'skipping run for runner ' \
                      f'{runner.short_uuid}'
                self._logger.info(msg)
                print(msg)
                continue
            runner.run(self.function.object, **run_args)

    def _run_direct_remote(self, run_args):

        raise NotImplementedError

    def _check_for_resultfiles(self) -> dict:
        self._logger.info('checking for finished runs')
        files_to_check = []
        for runner in self.runner_list:
            files_to_check.append(runner.resultpath)

        return self.url.utils.file_mtime(files_to_check)

    def fetch_results(self,
                      raise_if_not_finished: bool = False) -> list:
        """
        Collect any scripted run resultfiles and insert them into their runners

        Args:
            raise_if_not_finished (bool):
                raise an error if all calculations not finished

        Returns:
            None
        """
        present_runfiles = self._check_for_resultfiles()

        if not any(present_runfiles.values()):
            self._logger.info('no valid results found, exiting early')
            return [None]*len(self.runner_list)

        self._logger.info('present result files:')
        self._logger.info(format_iterable(present_runfiles))
        for runner in self.runner_list:
            if present_runfiles[runner.resultpath]:
                self.transport.queue_for_pull(os.path.split(
                                              runner.resultpath)[1],
                                              runner.local_dir,
                                              runner.run_dir)

        self._logger.info('pulling completed result files')
        self.transport.transfer()
        for runner in self.runner_list:
            pulled = runner.local_resultpath
            with open(pulled, 'r') as o:
                # TODO convert to full serialisation
                result = yaml.safe_load(o)

            timestamp = int(os.path.getmtime(pulled))

            if timestamp < runner.last_updated:
                self._logger.warning('calculation not completed yet')
                continue

            mtime = datetime.fromtimestamp(timestamp)
            runner.insert_history(mtime, 'resultfile created remotely')

            runner.result = result

        self.database.update(self.pack())

        if not self.all_finished and raise_if_not_finished:
            raise RuntimeError('Calculations not yet completed!')

    @property
    def is_finished(self):
        if self.skip:
            ret = []
            for runner in self.runner_list:
                if runner.is_finished:
                    ret.append(True)
                    continue
                ret.append(False)

        if self.script:
            self._logger.info('scripted run, checking for files')
            # look for the resultfiles
            fin = self._check_for_resultfiles()

            # create a list of the resultfiles that are available
            ret =  []
            for runner in self.runner_list:

                resultpath = runner.resultpath
                last_updated = runner.last_updated
                mtime = fin[resultpath]

                self._logger.debug(f'checking file {resultpath}. mtime '
                                   f'{mtime} vs runner time {last_updated}')

                if mtime is None:
                    ret.append(False)
                elif mtime >= last_updated:
                    ret.append(True)
                else:
                    ret.append(False)

            return ret

        return [r.is_finished for r in self.runner_list]

    @property
    def all_finished(self):
        return all(self.is_finished)

    @property
    def results(self):
        results = [r.result for r in self.runner_list]
        # if all([r is None for r in results]):
        #     self._logger.info('results are all None, fetching results')
        #     self.fetch_results()
        #     return self.results
        return results

    def clear_results(self):
        for runner in self.runner_list:
            file = runner.resultpath
            self._logger.info(f'attempting to clear result file {file}')
            try:
                os.remove(file)
                self._logger.info('Done')
            except FileNotFoundError:
                self._logger.info('file not found')



import collections.abc
import importlib
import inspect
import logging
import os.path

import yaml

CLASS_STORAGE_KEY = '~serialisedclass~'

_logger = logging.getLogger(__name__)


class SendableMixin:
    """
    baseclass for a "sendable" object, providing methods for conversion to
    yaml format
    """

    def pack(self,
             uuid: str = None,
             file: str = None):
        """
        "packs" the object into a dict-format, ready for yaml-dump

        Args:
            uuid (str):
                optional uuid string to package this data inside
            file (str):
                package to file

        Returns:
            (dict):
                packaged Runner
        """
        # remove any attributes not to be packaged
        # these objects are either not needed, or can be re-created on the
        # recipient end
        do_not_package = ['_logger',
                          '_transport',
                          '_database',
                          '_url',
                          ]

        payload = {k: self.serialise(v)
                   for k, v in self.__dict__.items()
                   if k not in do_not_package}

        payload[CLASS_STORAGE_KEY] = get_class_storage(self)

        if uuid:

            if '_uuid' in payload and payload['_uuid'] != uuid:
                raise ValueError('passed uuid and _uuid key in payload do'
                                 'not match!')

            payload['_uuid'] = uuid

            payload = {uuid: payload}

        if file is not None:
            print(f'dumping payload to {file}')
            with open(file, 'w+') as o:
                yaml.dump(payload, o)

            return None

        return payload

    @classmethod
    def unpack(cls,
               data: dict = None,
               file: str = None,
               limit: bool = True):
        """
        Re-create an object from a packaged payload coming from obj.pack

        .. note ::

            use this function to unpack from a payload _outside_ an object

        If the object takes a `dummy` flag, this will be inserted into the args
        The `dummy` kwarg MUST be in the arguments for a class which cannot be
        simply initialised with blank args, and logic within that object's
        __init__ to account for this, something along the lines of

        >>> def __init__(self, dummy):
        >>>     if dummy:
        >>>         return

        will break the object creation at that point

        Args:
            data (dict):
                __dict__ payload from the object that was packaged
            file (str):
                filepath to unpack from, if data is not given
            limit (bool):
                disable to allow outside classes to be unserialised
        Returns: re-created object
        """

        if data is None:
            if file is None:
                raise ValueError('please provide a data payload or filepath '
                                 'to read from')
            elif not os.path.isfile(file):
                raise FileNotFoundError(f'could not find file {file}')

            with open(file, 'r') as o:
                data = yaml.safe_load(o)

        # if the data passed is a {uuid: {data}} format dict, search for a uuid
        # within and extract the {data}
        firstkey = list(data)[0]
        if data[firstkey] is not None \
                and '_uuid' in data[firstkey] \
                and data[firstkey]['_uuid'] == firstkey:
            data = data[firstkey]

        # grab the required arguments for the class, and fill with placeholders
        # of the correct type
        params = inspect.signature(cls).parameters
        required_args = {}
        for arg_name, v in params.items():
            if v.default is inspect._empty and arg_name != 'kwargs':
                required_args[arg_name] = v.annotation

        if 'dummy' in params.keys():
            required_args['dummy'] = True

        # create a new instance of this class, using the placeholder args
        new = cls(**required_args)
        if hasattr(new, '_logger'):
            new._logger.info('blank class created with args')
        # update this object with the values extracted from the payload
        if isinstance(data, str):
            _logger.debug(f'importing from filename {data}')
            data = yaml.safe_load(data)

        # this is REQUIRED to be separate to the return call
        new.inject_payload(data, limit)

        return new

    def inject_payload(self,
                       payload: dict,
                       limit: bool = True):
        """
        inject payload into the __dict__, effectively re-creating the object

        .. note ::

            use this function to unpack _within_ an object

        Args:
            payload (dict):
                __dict__ payload from the object that was packaged
            limit (bool):
                enable or disable safeties
        """
        if hasattr(self, '_logger'):
            self._logger.info(f'finalising unpacking of {type(self)}')

        # temporary attribute to check for a valid class
        self._unpack_validate = True

        if CLASS_STORAGE_KEY in payload:
            selfkey = get_class_storage(self)
            packkey = payload[CLASS_STORAGE_KEY]

            if selfkey != packkey:
                raise RuntimeError(f'attempting to unpack class '
                                   f'{packkey["name"]} as '
                                   f'{selfkey["name"]}')

            delattr(self, '_unpack_validate')

        unpacked = {k: self.unserialise(v, limit) for k, v in payload.items()}

        self.__dict__.update(unpacked)

    def serialise(self,
                  obj):
        """
        Recurse over any iterable objects, or call the pack() method of any
        `SendableMixin` objects, for serialisation
        Args:
            obj:
                object to be packaged
        Returns:
            yaml-friendly object
        """
        if issubclass(type(obj), SendableMixin):

            payload = obj.pack()

            # add a passthrough for the object to be imported and recreated
            payload[CLASS_STORAGE_KEY] = get_class_storage(obj)
            return payload

        elif isinstance(obj, collections.abc.Mapping):
            # dict type
            return {k: self.serialise(v) for k, v in obj.items()}

        elif hasattr(obj, '__setitem__'):
            # list-type
            return [self.serialise(v) for v in obj]

        return obj

    def unserialise(self,
                    obj,
                    limit: bool = True):
        """
        Undo a serialised packaging, by importing the called object and calling
        its unpack() method
        Args:
            obj:
                payload to be unpacked
            limit (bool):
                exit if the object to be unpacked is not part of the
                remotemanager package
        Returns:
            object before packaging
        """
        try:
            unpackable = obj.get(CLASS_STORAGE_KEY)
        except AttributeError:
            unpackable = False

        if unpackable:
            # print(f'unserialising unpackable object {obj}')
            # extract the class to import
            source = obj.pop(CLASS_STORAGE_KEY)

            # import the module
            modulename = source['mod']
            if limit and not modulename.startswith('remotemanager'):
                raise ValueError('module to import is not within the '
                                 'remotemanager package, exiting for safety')
            mod = importlib.import_module(modulename)

            # now get the actual class to import and unpack
            cls = getattr(mod, source['name'])

            return cls.unpack(obj)

        elif isinstance(obj, collections.abc.Mapping):
            # dict type
            return {k: self.unserialise(v) for k, v in obj.items()}

        elif hasattr(obj, '__setitem__'):
            # list-type
            return [self.unserialise(v) for v in obj]

        return obj


def get_class_storage(obj):
    return {'mod': obj.__module__,
            'name': obj.__class__.__name__}



import functools
import logging
import operator
import yaml

from remotemanager.utils import get_version, ensure_filetype, \
    recursive_dict_update

from remotemanager.logging.utils import format_iterable


class Database:
    """
    Database file for use in the Dataset.
    Does not need to be called by the user

    Args:
        file (str):
            filename to write to
    """

    _versionkey = '~database-version~'

    def __init__(self,
                 file):

        file = ensure_filetype(file, 'yaml')

        self._logger = logging.getLogger(__name__ + '.Database')

        self._path = file
        self._storage = self._read()
        self._tree = None

    def _read(self):
        self._tree = None  # wipe tree cache
        self._logger.info(f'reading {self._path}')
        try:
            with open(self._path, 'r') as o:
                data = yaml.safe_load(o)
        except FileNotFoundError:
            self._logger.warning('file not found, creating')
            data = {Database._versionkey: get_version()}
            self._write(data)

        try:
            version = data.pop(Database._versionkey)
        except KeyError:
            version = "0"

        self._logger.info(f'database file version: {version}')

        return data

    def _write(self, data):
        self._tree = None  # wipe tree cache

        with open(self._path, 'w+') as o:
            yaml.dump(data, o)

    def read(self):
        self._storage = self._read()

    def write(self):
        self._write(self._storage)

    def update(self, payload):

        self._logger.info('updating stored info')

        self._storage = recursive_dict_update(self._storage, payload)
        self.write()

    @property
    def path(self):
        return self._path

    @property
    def tree(self):
        if self._tree is not None:
            return self._tree
        self._tree = self.climb(self._storage)
        return self._tree

    def climb(self,
              data: dict,
              branch: list = None) -> list:
        """
        "climb" a dictionary, returning a list of paths for each element

        Args:
            data (dict):
                dictionary to treat
            branch (list):
                current branch, used for recursion

        Returns:
            list of path-like strings

        """

        if branch is None:
            branch = []

        joinchar = '/'

        ret = []

        if len(data) == 0:
            ret.append(joinchar.join(branch))

        for key in data:
            tmp = []
            if isinstance(data[key], dict):
                tmp.append(key)
                ret += self.climb(data[key], branch + tmp)
            else:
                ret.append(joinchar.join(branch + [key]))
        return ret

    def find(self,
             key: str):
        """
        Find the first instance of key within the database tree
        Args:
            key (str):
                key (uuid) to look for
        Returns:
            database tree below key
        """

        def get_minimum_path(p, k):
            fullpath = p.split('/')
            return fullpath[:fullpath.index(k)+1]

        target = []
        for path in self.tree:
            if key in path:
                target = get_minimum_path(path, key)
                break
        data = chain_get(self._storage, target)

        return data


def chain_get(d: dict,
              keys: str):
    """
    get item from a nested dict using a list of keys
    Args:
        d (dict):
            nested dict to query
        keys (list):
            list of keys to use

    Returns:
        item from dict d
    """
    return functools.reduce(operator.getitem, keys, d)



import os
import math
import time

# need __version__, is there a better way than importing the whole package?
import remotemanager


def ensure_list(inp=None):
    if inp is None:
        return []
    elif isinstance(inp, (list, tuple, set)):
        return list(inp)
    elif isinstance(inp, str):
        return [inp]
    return list(inp)


def ensure_filetype(file, target_type):
    """
    Ensure that `file` is of type `type`
    Args:
        file (str):
            filename
        target_type (str):
            filetype to enforce

    Returns (str):
        filename of type `type`
    """
    fname, ftype = os.path.splitext(file)

    target_type = target_type.strip('.')

    return f'{fname}.{target_type}'


def safe_divide(a, b):
    """
    always-fit division. Rounds up after division, always returns >= 1
    """
    if b == 0:
        return 1
    r = math.ceil(a / b)
    return max(r, 1)


def get_version():
    return remotemanager.__version__


def recursive_dict_update(d1: dict,
                          d2: dict) -> dict:
    """
    Update d1 with all the keys of d2
    Args:
        d1 (dict):
            dictionary to be updated
        d2 (dict):
            dictionary to update with

    Returns (dict):
        updated d1

    """

    for k, v in d2.items():
        if isinstance(v, dict) and k in d1:
            d1[k] = recursive_dict_update(d1[k], d2[k])
        else:
            d1[k] = v

    return d1


def reverse_index(inplist, term):
    """
    return index of last occurrence of `term` in `inplist`
    Args:
        inplist (list):
            list to index
        term:
            object to index

    Returns (int):
        forward index of item

    """
    return len(inplist) - inplist[::-1].index(term) - 1


def integer_time_wait(offset: float = 0.0) -> None:
    """
    wait for an integer unix time + offset

    This function exists to increase the reproducibility of the tests

    Args:
        offset (float):
            extra offset to wait for
    """

    t0 = time.time()

    wait = 1 - math.fmod(t0, 1)

    time.sleep(wait + offset)

    return
